from ..Class.itemClass import KH2Item
from ..List.inventory import (
    ability,
    form,
    growth,
    magic,
    misc,
    proof,
    storyunlock,
    summon,
    synth,
    keyblade,
)
from ..List.inventory.form import DriveForm
from ..List.inventory.growth import GrowthType
from ..List.inventory.item import InventoryItem
from ..List.location.graph import RequirementFunction


class ItemPlacementHelpers:

    @staticmethod
    def need_fire_blizzard_thunder(inventory: list[int]) -> bool:
        return (
            magic.Fire.id in inventory
            and magic.Blizzard.id in inventory
            and magic.Thunder.id in inventory
        )

    @staticmethod
    def need_1_magnet(inventory: list[int]) -> bool:
        return inventory.count(magic.Magnet.id) >= 1

    @staticmethod
    def need_2_magnet(inventory: list[int]) -> bool:
        return inventory.count(magic.Magnet.id) >= 2

    @staticmethod
    def need_3_thunders(inventory: list[int]) -> bool:
        return inventory.count(magic.Thunder.id) == 3

    @staticmethod
    def count_growth(inventory: list[int], growth_type: GrowthType) -> int:
        ids = [
            item.id for item in growth.all_growth() if item.growth_type == growth_type
        ]
        count = 0
        for item in inventory:
            if item in ids:
                count += 1
        return count

    @staticmethod
    def need_growths(inventory: list[int]) -> bool:
        return (
            ItemPlacementHelpers.count_growth(inventory, GrowthType.HIGH_JUMP) >= 3
            and ItemPlacementHelpers.count_growth(inventory, GrowthType.QUICK_RUN) >= 3
            and ItemPlacementHelpers.count_growth(inventory, GrowthType.AERIAL_DODGE)
            >= 3
            and ItemPlacementHelpers.count_growth(inventory, GrowthType.GLIDE) >= 3
        )

    @staticmethod
    def need_proof_connection(inventory: list[int]) -> bool:
        return proof.ProofOfConnection.id in inventory

    @staticmethod
    def need_proof_peace(inventory: list[int]) -> bool:
        return proof.ProofOfPeace.id in inventory

    @staticmethod
    def has_valor_form(inventory: list[int]) -> bool:
        return form.ValorForm.id in inventory

    @staticmethod
    def has_wisdom_form(inventory: list[int]) -> bool:
        return form.WisdomForm.id in inventory

    @staticmethod
    def has_limit_form(inventory: list[int]) -> bool:
        return form.LimitForm.id in inventory

    @staticmethod
    def has_master_form(inventory: list[int]) -> bool:
        return form.MasterForm.id in inventory

    @staticmethod
    def has_final_form(inventory: list[int]) -> bool:
        return form.FinalForm.id in inventory

    @staticmethod
    def can_level_valor(inventory: list[int]) -> bool:
        return True

    @staticmethod
    def can_level_wisdom(inventory: list[int]) -> bool:
        return True

    @staticmethod
    def can_level_limit(inventory: list[int]) -> bool:
        return True

    @staticmethod
    def can_level_master(inventory: list[int]) -> bool:
        return ItemPlacementHelpers.ht1_check(
            inventory
        ) or ItemPlacementHelpers.hb2_check(inventory)

    @staticmethod
    def can_level_final(inventory: list[int]) -> bool:
        return (
            ItemPlacementHelpers.twtnw_roxas_check(inventory)
            or ItemPlacementHelpers.pr2_check(inventory)
            or ItemPlacementHelpers.bc2_check(inventory)
            or ItemPlacementHelpers.lod2_check(inventory)
            or ItemPlacementHelpers.tt3_check(inventory)
        )

    @staticmethod
    def has_auto_valor(inventory: list[int]) -> bool:
        return ability.AutoValor.id in inventory

    @staticmethod
    def has_auto_wisdom(inventory: list[int]) -> bool:
        return ability.AutoWisdom.id in inventory

    @staticmethod
    def has_auto_limit(inventory: list[int]) -> bool:
        return ability.AutoLimitForm.id in inventory

    @staticmethod
    def has_auto_master(inventory: list[int]) -> bool:
        return ability.AutoMaster.id in inventory

    @staticmethod
    def has_auto_final(inventory: list[int]) -> bool:
        return ability.AutoFinal.id in inventory

    @staticmethod
    def count_forms(inventory: list[int]) -> int:
        count = 0
        if ItemPlacementHelpers.has_valor_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_wisdom_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_limit_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_master_form(inventory):
            count += 1
        if ItemPlacementHelpers.has_final_form(inventory):
            count += 1
        return count

    @staticmethod
    def need_forms(inventory: list[int]) -> bool:
        return ItemPlacementHelpers.count_forms(inventory) == 5

    @staticmethod
    def need_summons(inventory: list[int]) -> bool:
        return (
            summon.LampCharm.id in inventory
            and summon.FeatherCharm.id in inventory
            and summon.UkuleleCharm.id in inventory
            and summon.BaseballCharm.id in inventory
        )

    @staticmethod
    def count_pages(inventory: list[int]) -> int:
        return inventory.count(misc.TornPages.id)

    @staticmethod
    def need_torn_pages(count: int) -> RequirementFunction:
        return lambda inventory: ItemPlacementHelpers.count_pages(inventory) >= count

    @staticmethod
    def need_proofs(inventory: list[int]) -> bool:
        return (
            proof.ProofOfConnection.id in inventory
            and proof.ProofOfNonexistence.id in inventory
            and proof.ProofOfPeace.id in inventory
        )

    @staticmethod
    def need_promise_charm(inventory: list[int]) -> bool:
        return misc.PromiseCharm.id in inventory

    @staticmethod
    def make_need_objectives_lambda(num_objectives_needed: int):
        return (
            lambda inventory: inventory.count(misc.ObjectiveItem.id)
            >= num_objectives_needed
        )

    @staticmethod
    def make_form_lambda(drive_form: DriveForm, form_level: int) -> RequirementFunction:
        if drive_form == form.ValorForm:
            return (
                lambda inventory: ItemPlacementHelpers.has_valor_form(inventory)
                and ItemPlacementHelpers.can_level_valor(inventory)
                and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
            )
        if drive_form == form.WisdomForm:
            return (
                lambda inventory: ItemPlacementHelpers.has_wisdom_form(inventory)
                and ItemPlacementHelpers.can_level_wisdom(inventory)
                and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
            )
        if drive_form == form.LimitForm:
            return (
                lambda inventory: ItemPlacementHelpers.has_limit_form(inventory)
                and ItemPlacementHelpers.can_level_limit(inventory)
                and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
            )
        if drive_form == form.MasterForm:
            return (
                lambda inventory: ItemPlacementHelpers.has_master_form(inventory)
                and ItemPlacementHelpers.can_level_master(inventory)
                and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
            )
        if drive_form == form.FinalForm:
            return (
                lambda inventory: ItemPlacementHelpers.has_final_form(inventory)
                and ItemPlacementHelpers.can_level_final(inventory)
                and ItemPlacementHelpers.count_forms(inventory) >= form_level - 2
            )
        return lambda inventory: False

    @staticmethod
    def make_form_lambda_nightmare(
        drive_form: DriveForm, form_level: int
    ) -> RequirementFunction:
        def count_auto_forms(inventory: list[int]) -> int:
            count = 0
            if ItemPlacementHelpers.has_auto_valor(inventory):
                count += 1
            if ItemPlacementHelpers.has_auto_wisdom(inventory):
                count += 1
            if ItemPlacementHelpers.has_auto_limit(inventory):
                count += 1
            if ItemPlacementHelpers.has_auto_master(inventory):
                count += 1
            return count

        def final_possible_but_not_obtained(inventory: list[int]) -> bool:
            return (
                ItemPlacementHelpers.has_valor_form(inventory)
                or ItemPlacementHelpers.has_wisdom_form(inventory)
                or ItemPlacementHelpers.has_limit_form(inventory)
                or ItemPlacementHelpers.has_master_form(inventory)
                or count_auto_forms(inventory) >= 1
            ) and not ItemPlacementHelpers.has_final_form(inventory)

        def max_form_level(inventory: list[int]) -> int:
            base_count = ItemPlacementHelpers.count_forms(inventory) + 2
            if final_possible_but_not_obtained(inventory):
                return base_count + 1
            else:
                return base_count

        if drive_form == form.ValorForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_valor_form(inventory)
                    or ItemPlacementHelpers.has_auto_valor(inventory)
                )
                and ItemPlacementHelpers.can_level_valor(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.WisdomForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_wisdom_form(inventory)
                    or ItemPlacementHelpers.has_auto_wisdom(inventory)
                )
                and ItemPlacementHelpers.can_level_wisdom(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.LimitForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_limit_form(inventory)
                    or ItemPlacementHelpers.has_auto_limit(inventory)
                )
                and ItemPlacementHelpers.can_level_limit(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.MasterForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_master_form(inventory)
                    or ItemPlacementHelpers.has_auto_master(inventory)
                )
                and ItemPlacementHelpers.can_level_master(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.FinalForm:

            def have_final_form(inventory: list[int]) -> bool:
                return ItemPlacementHelpers.has_final_form(
                    inventory
                ) or final_possible_but_not_obtained(inventory)

            if form_level == 2:
                return lambda inventory: (
                    have_final_form(inventory)
                    or ItemPlacementHelpers.has_auto_final(inventory)
                ) and ItemPlacementHelpers.can_level_final(inventory)
            else:
                return (
                    lambda inventory: have_final_form(inventory)
                    and max_form_level(inventory) >= form_level
                    and ItemPlacementHelpers.can_level_final(inventory)
                )

        return lambda inventory: False

    @staticmethod
    def make_form_lambda_nightmare_no_anti(
        drive_form: DriveForm, form_level: int
    ) -> RequirementFunction:
        def max_form_level(inventory: list[int]) -> int:
            base_count = ItemPlacementHelpers.count_forms(inventory) + 2
            return base_count

        if drive_form == form.ValorForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_valor_form(inventory)
                    or ItemPlacementHelpers.has_auto_valor(inventory)
                )
                and ItemPlacementHelpers.can_level_valor(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.WisdomForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_wisdom_form(inventory)
                    or ItemPlacementHelpers.has_auto_wisdom(inventory)
                )
                and ItemPlacementHelpers.can_level_wisdom(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.LimitForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_limit_form(inventory)
                    or ItemPlacementHelpers.has_auto_limit(inventory)
                )
                and ItemPlacementHelpers.can_level_limit(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.MasterForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_master_form(inventory)
                    or ItemPlacementHelpers.has_auto_master(inventory)
                )
                and ItemPlacementHelpers.can_level_master(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.FinalForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_final_form(inventory)
                    or ItemPlacementHelpers.has_auto_final(inventory)
                )
                and ItemPlacementHelpers.can_level_final(inventory)
                and max_form_level(inventory) >= form_level
            )

        return lambda inventory: False

    @staticmethod
    def make_form_lambda_nightmare_no_final(
        drive_form: DriveForm, form_level: int
    ) -> RequirementFunction:
        def final_possible_but_not_obtained(_: list[int]) -> bool:
            return False

        def max_form_level(inventory: list[int]) -> int:
            base_count = ItemPlacementHelpers.count_forms(inventory) + 2
            if final_possible_but_not_obtained(inventory):
                return base_count + 1
            else:
                return base_count

        if drive_form == form.ValorForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_valor_form(inventory)
                    or ItemPlacementHelpers.has_auto_valor(inventory)
                )
                and ItemPlacementHelpers.can_level_valor(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.WisdomForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_wisdom_form(inventory)
                    or ItemPlacementHelpers.has_auto_wisdom(inventory)
                )
                and ItemPlacementHelpers.can_level_wisdom(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.LimitForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_limit_form(inventory)
                    or ItemPlacementHelpers.has_auto_limit(inventory)
                )
                and ItemPlacementHelpers.can_level_limit(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.MasterForm:
            return (
                lambda inventory: (
                    ItemPlacementHelpers.has_master_form(inventory)
                    or ItemPlacementHelpers.has_auto_master(inventory)
                )
                and ItemPlacementHelpers.can_level_master(inventory)
                and max_form_level(inventory) >= form_level
            )
        if drive_form == form.FinalForm:

            def have_final_form(inventory: list[int]) -> bool:
                return final_possible_but_not_obtained(
                    inventory
                ) or ItemPlacementHelpers.has_final_form(inventory)

            if form_level == 2:
                return lambda inventory: (
                    have_final_form(inventory)
                    or ItemPlacementHelpers.has_auto_final(inventory)
                ) and ItemPlacementHelpers.can_level_final(inventory)
            else:
                return (
                    lambda inventory: have_final_form(inventory)
                    and max_form_level(inventory) >= form_level
                    and ItemPlacementHelpers.can_level_final(inventory)
                )

        return lambda inventory: False

    @staticmethod
    def get_number_visit_unlocks(inventory: list[int]) -> int:
        visit_unlock_ids = [u.id for u in storyunlock.all_story_unlocks()]
        running_total = 0
        for id in visit_unlock_ids:
            running_total += inventory.count(id)
        return running_total

    @staticmethod
    def make_level_group_check(group_index) -> RequirementFunction:
        thresholds = [
            1,
            3,
            6,
            9,
            12,
            15,
            18,
            21,
            24,
        ]  # modified from AP so that max level could in theory have a visit unlock
        if group_index < len(thresholds):
            return (
                lambda inv: ItemPlacementHelpers.get_number_visit_unlocks(inv)
                >= thresholds[group_index]
            )
        else:
            return (
                lambda inv: ItemPlacementHelpers.get_number_visit_unlocks(inv)
                >= thresholds[-1]
            )

    @staticmethod
    def stt_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.NaminesSketches.id) == 1

    @staticmethod
    def dc1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.RoyalSummons.id) >= 1

    @staticmethod
    def dc2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.RoyalSummons.id) == 2

    @staticmethod
    def oc1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BattlefieldsOfWar.id) >= 1

    @staticmethod
    def oc2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BattlefieldsOfWar.id) == 2

    @staticmethod
    def lod1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SwordOfTheAncestor.id) >= 1

    @staticmethod
    def lod2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SwordOfTheAncestor.id) == 2

    @staticmethod
    def bc1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BeastsClaw.id) >= 1

    @staticmethod
    def bc2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BeastsClaw.id) == 2

    @staticmethod
    def ht1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BoneFist.id) >= 1

    @staticmethod
    def ht2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.BoneFist.id) == 2

    @staticmethod
    def pl1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.ProudFang.id) >= 1

    @staticmethod
    def pl2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.ProudFang.id) == 2

    @staticmethod
    def pr1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SkillAndCrossbones.id) >= 1

    @staticmethod
    def pr2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.SkillAndCrossbones.id) == 2

    @staticmethod
    def ag1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.Scimitar.id) >= 1

    @staticmethod
    def ag2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.Scimitar.id) == 2

    @staticmethod
    def twtnw_roxas_check(inventory: list[int]) -> bool:
        return (
            inventory.count(storyunlock.WayToTheDawn.id) >= 1
            and ItemPlacementHelpers.get_number_visit_unlocks(inventory) > 10
        )

    @staticmethod
    def twtnw_post_saix_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.WayToTheDawn.id) == 2

    @staticmethod
    def sp1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IdentityDisk.id) >= 1

    @staticmethod
    def sp2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IdentityDisk.id) == 2

    @staticmethod
    def tt1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IceCream.id) >= 1

    @staticmethod
    def tt2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IceCream.id) >= 2

    @staticmethod
    def tt3_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.IceCream.id) == 3

    @staticmethod
    def hb1_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.MembershipCard.id) >= 1

    @staticmethod
    def hb2_check(inventory: list[int]) -> bool:
        return inventory.count(storyunlock.MembershipCard.id) == 2

    @staticmethod
    def make_synth_requirement(synth_item: KH2Item) -> RequirementFunction:
        # making lambdas for the different synth items
        default_access_lambda: RequirementFunction = lambda inventory: True
        synth_req_map: dict[InventoryItem, RequirementFunction] = {
            synth.BlazingShard: lambda inv: ItemPlacementHelpers.dc2_check(inv)
            or ItemPlacementHelpers.bc2_check(inv)
            or ItemPlacementHelpers.oc2_check(inv),
            synth.BlazingStone: lambda inv: ItemPlacementHelpers.lod2_check(inv)
            or ItemPlacementHelpers.sp2_check(inv)
            or ItemPlacementHelpers.oc2_check(inv),
            synth.BlazingGem: ItemPlacementHelpers.ag1_check,
            synth.BlazingCrystal: lambda inv: ItemPlacementHelpers.oc2_check(inv)
            or ItemPlacementHelpers.hb2_check(inv)
            or ItemPlacementHelpers.bc2_check(inv)
            or ItemPlacementHelpers.ag2_check(inv),
            # shadows are guaranteed anytime in DC/TR
            synth.DarkShard: ItemPlacementHelpers.dc1_check,
            # LoD1 and 2 both have them
            synth.DarkStone: ItemPlacementHelpers.lod1_check,
            # BC1 and 2 have them
            synth.DarkGem: ItemPlacementHelpers.bc1_check,
            # PR1 has them, but could be locked out
            synth.DarkCrystal: lambda inv: ItemPlacementHelpers.lod2_check(inv),
            # TWTNW has them post Xemnas
            synth.DenseShard: lambda inv: ItemPlacementHelpers.twtnw_post_saix_check(
                inv
            )
            or ItemPlacementHelpers.bc2_check(inv)
            or ItemPlacementHelpers.tt3_check(inv),
            # TWTNW has them post Xemnas
            synth.DenseStone: lambda inv: ItemPlacementHelpers.lod2_check(inv)
            or ItemPlacementHelpers.twtnw_roxas_check(inv),
            # TWTNW has them post Xemnas
            synth.DenseGem: lambda inv: ItemPlacementHelpers.tt3_check(inv)
            or ItemPlacementHelpers.stt_check(inv)
            or ItemPlacementHelpers.twtnw_roxas_check(inv),
            # TWTNW has them post Xemnas
            synth.DenseCrystal: lambda inv: ItemPlacementHelpers.tt3_check(inv)
            or ItemPlacementHelpers.twtnw_roxas_check(inv)
            or ItemPlacementHelpers.pl2_check(inv),
            # BC ones can disappear
            synth.FrostShard: lambda inv: ItemPlacementHelpers.lod2_check(inv)
            or ItemPlacementHelpers.ag2_check(inv)
            or ItemPlacementHelpers.pr2_check(inv),
            # DC/TR
            synth.FrostStone: ItemPlacementHelpers.dc2_check,
            # AG1 and 2 both have them
            synth.FrostGem: ItemPlacementHelpers.ag1_check,
            # PL1 and 2 both have them
            synth.FrostCrystal: ItemPlacementHelpers.pl1_check,
            # LoD1 and 2 both have them
            synth.LightningShard: lambda inv: ItemPlacementHelpers.lod1_check(inv)
            or ItemPlacementHelpers.dc1_check(inv)
            or ItemPlacementHelpers.pl2_check(inv)
            or ItemPlacementHelpers.pr2_check(inv)
            or ItemPlacementHelpers.ag2_check(inv),
            # HT and SP ones can become unavailable
            synth.LightningStone: lambda inv: ItemPlacementHelpers.oc2_check(inv)
            or ItemPlacementHelpers.lod2_check(inv)
            or ItemPlacementHelpers.ag2_check(inv),
            synth.LightningGem: lambda inv: ItemPlacementHelpers.hb2_check(inv)
            or ItemPlacementHelpers.bc2_check(inv)
            or ItemPlacementHelpers.pr2_check(inv)
            or ItemPlacementHelpers.ht2_check(inv)
            or ItemPlacementHelpers.pl2_check(inv)
            or ItemPlacementHelpers.sp2_check(inv),
            # SP1 and 2 both have them
            synth.LightningCrystal: ItemPlacementHelpers.sp1_check,
            # OC1 can be locked out
            synth.LucidShard: lambda inv: ItemPlacementHelpers.pr2_check(inv)
            or ItemPlacementHelpers.ht2_check(inv),
            # HT1 and 2 have them
            synth.LucidStone: ItemPlacementHelpers.ht1_check,
            # SP1 and 2 both have them
            synth.LucidGem: lambda inv: ItemPlacementHelpers.hb2_check(inv)
            or ItemPlacementHelpers.sp1_check(inv),
            synth.LucidCrystal: lambda inv: ItemPlacementHelpers.lod2_check(inv)
            or ItemPlacementHelpers.bc2_check(inv)
            or ItemPlacementHelpers.oc2_check(inv)
            or ItemPlacementHelpers.pr2_check(inv)
            or ItemPlacementHelpers.ag2_check(inv)
            or ItemPlacementHelpers.ht2_check(inv)
            or ItemPlacementHelpers.pl2_check(inv)
            or ItemPlacementHelpers.hb2_check(inv),
            # BC1 and 2 both have them, also TR
            synth.PowerShard: lambda inv: ItemPlacementHelpers.bc1_check(inv)
            or ItemPlacementHelpers.dc2_check(inv)
            or ItemPlacementHelpers.ht2_check(inv),
            # AG1 and 2 both have them
            synth.PowerStone: ItemPlacementHelpers.ag1_check,
            # PL1 and 2 both have them
            synth.PowerGem: ItemPlacementHelpers.pl1_check,
            synth.PowerCrystal: lambda inv: ItemPlacementHelpers.hb2_check(inv)
            or ItemPlacementHelpers.bc2_check(inv)
            or ItemPlacementHelpers.oc2_check(inv)
            or ItemPlacementHelpers.pr2_check(inv),
            # TWTNW always has them in Luxord and Saix rooms
            synth.TwilightShard: lambda inv: ItemPlacementHelpers.twtnw_roxas_check(inv)
            or ItemPlacementHelpers.pr2_check(inv)
            or ItemPlacementHelpers.tt3_check(inv)
            or ItemPlacementHelpers.lod2_check(inv),
            # TWTNW always has them
            synth.TwilightStone: lambda inv: ItemPlacementHelpers.tt3_check(inv)
            or ItemPlacementHelpers.twtnw_roxas_check(inv)
            or ItemPlacementHelpers.hb2_check(inv),
            # TWTNW always has them
            synth.TwilightGem: lambda inv: ItemPlacementHelpers.twtnw_roxas_check(inv)
            or ItemPlacementHelpers.tt3_check(inv),
            # TWTNW always has them
            synth.TwilightCrystal: ItemPlacementHelpers.twtnw_roxas_check,
        }

        item = synth_item.item
        if item in synth_req_map:
            return synth_req_map[item]

        # default returned lambda, doesn't allow for acquiring the item, failsafe
        return lambda inventory: False

    @staticmethod
    def need_stt_keyblade(inventory: list[int]) -> bool:
        return keyblade.BondOfFlame.id in inventory

    @staticmethod
    def need_tt_keyblade(inventory: list[int]) -> bool:
        return keyblade.Oathkeeper.id in inventory

    @staticmethod
    def need_hb_keyblade(inventory: list[int]) -> bool:
        return keyblade.SleepingLion.id in inventory

    @staticmethod
    def need_cor_keyblade(inventory: list[int]) -> bool:
        return keyblade.WinnersProof.id in inventory

    @staticmethod
    def need_lod_keyblade(inventory: list[int]) -> bool:
        return keyblade.HiddenDragon.id in inventory

    @staticmethod
    def need_bc_keyblade(inventory: list[int]) -> bool:
        return keyblade.RumblingRose.id in inventory

    @staticmethod
    def need_oc_keyblade(inventory: list[int]) -> bool:
        return keyblade.HerosCrest.id in inventory

    @staticmethod
    def need_dc_keyblade(inventory: list[int]) -> bool:
        return keyblade.Monochrome.id in inventory

    @staticmethod
    def need_pr_keyblade(inventory: list[int]) -> bool:
        return keyblade.FollowTheWind.id in inventory

    @staticmethod
    def need_ag_keyblade(inventory: list[int]) -> bool:
        return keyblade.WishingLamp.id in inventory

    @staticmethod
    def need_ht_keyblade(inventory: list[int]) -> bool:
        return keyblade.DecisivePumpkin.id in inventory

    @staticmethod
    def need_pl_keyblade(inventory: list[int]) -> bool:
        return keyblade.CircleOfLife.id in inventory

    @staticmethod
    def need_sp_keyblade(inventory: list[int]) -> bool:
        return keyblade.PhotonDebugger.id in inventory

    @staticmethod
    def need_twtnw_keyblade(inventory: list[int]) -> bool:
        return keyblade.TwoBecomeOne.id in inventory

    @staticmethod
    def need_haw_keyblade(inventory: list[int]) -> bool:
        return keyblade.SweetMemories.id in inventory



from typing import Optional

from ...Class.newLocationClass import KH2Location
from ..LvupStats import DreamWeaponOffsets
from ..configDict import locationCategory, locationType
from ..inventory import ability
from ..inventory.item import InventoryItem
from .graph import LocationGraphBuilder, START_NODE, RequirementEdge
from ...Module.itemPlacementRestriction import ItemPlacementHelpers


def level_reward(
    level: int, description: str, vanilla: Optional[InventoryItem]
) -> KH2Location:
    vanilla_items: list[int] = []
    if vanilla is not None:
        vanilla_items.append(vanilla.id)
    return KH2Location(
        level,
        description,
        locationCategory.LEVEL,
        [locationType.Level],
        VanillaItems=vanilla_items,
    )


def make_graph(graph: LocationGraphBuilder):
    settings = graph.settings
    max_level = settings.max_level_checks

    vanilla_levels: dict[int, dict[int, InventoryItem]] = {
        50: {
            2: ability.ComboBoost,
            4: ability.ExperienceBoost,
            7: ability.MagicLockOn,
            9: ability.ReactionBoost,
            10: ability.ItemBoost,
            12: ability.LeafBracer,
            14: ability.FireBoost,
            15: ability.DriveBoost,
            17: ability.Draw,
            20: ability.CombinationBoost,
            23: ability.DamageDrive,
            25: ability.AirComboBoost,
            28: ability.BlizzardBoost,
            30: ability.DriveConverter,
            32: ability.NegativeCombo,
            34: ability.OnceMore,
            36: ability.FinishingPlus,
            39: ability.ThunderBoost,
            41: ability.Defender,
            44: ability.BerserkCharge,
            46: ability.Jackpot,
            48: ability.SecondChance,
            50: ability.DamageControl,
        },
        99: {
            7: ability.ComboBoost,
            9: ability.ExperienceBoost,
            12: ability.MagicLockOn,
            15: ability.ReactionBoost,
            17: ability.ItemBoost,
            20: ability.LeafBracer,
            23: ability.FireBoost,
            25: ability.DriveBoost,
            28: ability.Draw,
            31: ability.CombinationBoost,
            33: ability.DamageDrive,
            36: ability.AirComboBoost,
            39: ability.BlizzardBoost,
            41: ability.DriveConverter,
            44: ability.NegativeCombo,
            47: ability.OnceMore,
            49: ability.FinishingPlus,
            53: ability.ThunderBoost,
            59: ability.Defender,
            65: ability.BerserkCharge,
            73: ability.Jackpot,
            85: ability.SecondChance,
            99: ability.DamageControl,
        },
    }

    node_index = 0
    locations: list[str] = []
    current_location_list: list[KH2Location] = []

    already_added_level_to_group = False
    level_offsets = DreamWeaponOffsets()
    excluded_levels = settings.excluded_levels()
    for level in range(1, 100):
        level_description = f"Level {level}"

        if settings.split_levels:
            shield = level_offsets.get_shield_level(max_level, level)
            staff = level_offsets.get_staff_level(max_level, level)
            if shield and staff:
                level_description = f"Level Sw: {level} Sh: {shield} St: {staff}"

        vanilla_item: Optional[InventoryItem] = None
        if max_level in vanilla_levels and level in vanilla_levels[max_level]:
            vanilla_item = vanilla_levels[max_level][level]
        elif level in vanilla_levels[99]:
            vanilla_item = vanilla_levels[99][level]

        current_location_list.append(
            level_reward(level, level_description, vanilla_item)
        )

        if level not in excluded_levels:
            if already_added_level_to_group:
                locations.append(
                    graph.add_location(
                        f"LevelGroup-{node_index}", current_location_list
                    )
                )
                current_location_list = []
                node_index += 1
                already_added_level_to_group = False
            else:
                already_added_level_to_group = True

    # Add in the last group if there are any stragglers
    if len(current_location_list) > 0:
        locations.append(
            graph.add_location(f"LevelGroup-{node_index}", current_location_list)
        )

    for index, location in enumerate(locations):
        if index == 0:
            graph.add_edge(
                START_NODE,
                location,
                RequirementEdge(req=ItemPlacementHelpers.make_level_group_check(index)),
            )
        else:
            graph.add_edge(
                locations[index - 1],
                location,
                RequirementEdge(req=ItemPlacementHelpers.make_level_group_check(index)),
            )



from enum import Enum

from ..configDict import locationType
from ..inventory import keyblade, magic
from .graph import (
    RequirementEdge,
    popup,
    LocationGraphBuilder,
    START_NODE,
    DefaultLogicGraph,
)
from ...Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    AtlanticaTutorial = "Atlantica Tutorial"
    Ursula = "Ursula"
    NewDayIsDawning = "New Day is Dawning"


class CheckLocation(str, Enum):
    UnderseaKingdomMap = "Undersea Kingdom Map"
    MysteriousAbyss = "Mysterious Abyss"
    MusicalBlizzardElement = "Musical Blizzard Element"
    MusicalOrichalcumPlus = "Musical Orichalcum+"


class ATLogicGraph(DefaultLogicGraph):
    def __init__(self, reverse_rando):
        DefaultLogicGraph.__init__(self, NodeId)
        if not reverse_rando:
            self.logic[NodeId.AtlanticaTutorial][
                NodeId.Ursula
            ] = ItemPlacementHelpers.need_2_magnet
            self.logic[NodeId.Ursula][
                NodeId.NewDayIsDawning
            ] = ItemPlacementHelpers.need_3_thunders
        else:
            self.logic[NodeId.NewDayIsDawning][
                NodeId.Ursula
            ] = ItemPlacementHelpers.need_1_magnet
            self.logic[NodeId.Ursula][NodeId.AtlanticaTutorial] = (
                lambda inv: ItemPlacementHelpers.need_2_magnet(inv)
                and ItemPlacementHelpers.need_3_thunders(inv)
            )


def make_graph(graph: LocationGraphBuilder):
    at = locationType.Atlantica
    at_logic = ATLogicGraph(graph.reverse_rando)
    graph.add_logic(at_logic)

    tutorial = graph.add_location(
        NodeId.AtlanticaTutorial,
        [
            popup(367, CheckLocation.UnderseaKingdomMap, at),
        ],
    )
    ursula = graph.add_location(
        NodeId.Ursula,
        [
            popup(
                287, CheckLocation.MysteriousAbyss, at, vanilla=keyblade.MysteriousAbyss
            ),
        ],
    )
    new_day_is_dawning = graph.add_location(
        NodeId.NewDayIsDawning,
        [
            popup(
                279, CheckLocation.MusicalBlizzardElement, at, vanilla=magic.Blizzard
            ),
            popup(538, CheckLocation.MusicalOrichalcumPlus, at),
        ],
    )

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, tutorial)
        graph.add_edge(tutorial, ursula, RequirementEdge())
        graph.add_edge(ursula, new_day_is_dawning, RequirementEdge())
    else:
        graph.add_edge(START_NODE, new_day_is_dawning)
        graph.add_edge(new_day_is_dawning, ursula, RequirementEdge())
        graph.add_edge(ursula, tutorial, RequirementEdge())



import json
import os
import random
from pathlib import Path

from ..Class import settingkey
from ..Class.openkhmod import Asset
from ..Class.seedSettings import SeedSettings
from . import appconfig
from .cosmeticsmods import music
from .cosmeticsmods.endingpic import EndingPictureRandomizer
from .cosmeticsmods.field2d import CommandMenuRandomizer, RoomTransitionImageRandomizer
from .cosmeticsmods.itempic import ItempicRandomizer
from .cosmeticsmods.keyblade import KeybladeRandomizer


class CustomCosmetics:

    def __init__(self):
        super().__init__()

        config_path = Path("auto-save") / "custom-cosmetics.json"
        self.config_path = config_path
        self.external_executables: list[str] = []

        if config_path.is_file():
            with open(config_path, encoding="utf-8") as config_file:
                try:
                    raw_json: dict = json.load(config_file)
                    self.external_executables = raw_json.get("external_executables", [])
                except Exception:
                    print(
                        "Error decoding custom-cosmetics.json, using empty list of external executables"
                    )

    def add_custom_executable(self, path_str: str):
        self.external_executables.append(path_str)

    def remove_executable_at_index(self, index: int):
        del self.external_executables[index]

    def write_file(self):
        with open(self.config_path, "w", encoding="utf-8") as config_file:
            raw_json = {"external_executables": self.external_executables}
            json.dump(raw_json, config_file, indent=4)

    def collect_custom_executable_files(self) -> list[str]:
        return [custom_file for custom_file in self.external_executables]


class CosmeticsMod:

    @staticmethod
    def bootstrap_cosmetics_files():
        CosmeticsMod.bootstrap_music_list_file()
        ItempicRandomizer.bootstrap_itempic_file()

        custom_music_path = appconfig.read_custom_music_path()
        if custom_music_path is not None:
            CosmeticsMod.bootstrap_custom_music_folder(custom_music_path)

        custom_visuals_path = appconfig.read_custom_visuals_path()
        if custom_visuals_path is not None:
            CosmeticsMod.bootstrap_custom_visuals_folder(custom_visuals_path)

    @staticmethod
    def bootstrap_music_list_file() -> Path:
        """Creates the musiclist file if needed."""
        music_list_file_path = Path("musiclist.json")
        if not music_list_file_path.is_file():
            with open(
                music_list_file_path, mode="w", encoding="utf-8"
            ) as music_list_file:
                json.dump(music.kh2_music_list, music_list_file, indent=4)
        return music_list_file_path

    @staticmethod
    def bootstrap_custom_music_folder(custom_music_path: Path):
        """Creates folders for each of the default music categories."""
        for folder in [
            "atlantica",
            "battle",
            "boss",
            "cutscene",
            "field",
            "title",
            "wild",
        ]:
            (custom_music_path / folder).mkdir(exist_ok=True)

    @staticmethod
    def randomize_music(
        ui_settings: SeedSettings,
    ) -> tuple[list[Asset], dict[str, str]]:
        """
        Randomizes music, returning a list of assets to be added to the seed mod and a dictionary of which song was
        replaced by which replacement.
        """
        return CosmeticsMod._get_music_assets(ui_settings)

    @staticmethod
    def get_keyblade_summary() -> dict[str, int]:
        vanilla = KeybladeRandomizer.collect_vanilla_keyblades()
        custom = KeybladeRandomizer.collect_custom_keyblades()
        return {"In-Game": len(vanilla), "Custom": len(custom)}

    @staticmethod
    def get_music_summary(settings: SeedSettings) -> dict[str, int]:
        music_files = CosmeticsMod._collect_music_files(settings)

        return {category: len(song_list) for category, song_list in music_files.items()}

    @staticmethod
    def _collect_music_files(settings: SeedSettings) -> dict[str, list[Path]]:
        """Returns music files grouped by category."""
        result: dict[str, list[Path]] = {}

        categorize = settings.get(settingkey.MUSIC_RANDO_PC_USE_CATEGORIES)
        dmca_safe = settings.get(settingkey.MUSIC_RANDO_PC_DMCA_SAFE)

        if settings.get(settingkey.MUSIC_RANDO_PC_INCLUDE_CUSTOM):
            custom_music_path = appconfig.read_custom_music_path()
            if custom_music_path is not None:
                for child in [
                    str(category_file).lower()
                    for category_file in os.listdir(custom_music_path)
                ]:
                    child_path = custom_music_path / child
                    if child_path.is_dir():
                        category_songs: list[Path] = []
                        for root, dirs, files in os.walk(child_path):
                            root_path = Path(root)
                            for file in files:
                                _, extension = os.path.splitext(file)
                                if extension.lower() == ".scd":
                                    file_path = root_path / file
                                    category_songs.append(file_path)
                        resolved_category = child
                        if not categorize:
                            resolved_category = "wild"
                        if resolved_category not in result:
                            result[resolved_category] = []
                        result[resolved_category] += category_songs

        extracted_data_path = appconfig.extracted_data_path()
        if extracted_data_path is not None:

            def add_game_song(song_file_path: Path, category: str, song_dmca: bool):
                if not song_file_path.is_file():
                    return

                if not categorize:
                    category = "wild"
                if category not in result:
                    result[category] = []
                if not dmca_safe or not song_dmca:
                    result[category].append(song_file_path)

            if settings.get(settingkey.MUSIC_RANDO_PC_INCLUDE_KH2):
                kh2_path = extracted_data_path / "kh2"
                if kh2_path.is_dir():
                    for kh2_song in music.kh2_music_list:
                        add_game_song(
                            song_file_path=kh2_path / kh2_song["filename"],
                            category=kh2_song["type"][0].lower(),
                            song_dmca=kh2_song.get("dmca", False),
                        )

            def add_other_game_music(
                enabled_key: str, game_music_path: Path, game_music_list: list
            ):
                if settings.get(enabled_key) and game_music_path.is_dir():
                    for song in game_music_list:
                        add_game_song(
                            song_file_path=game_music_path / song["name"],
                            category=song["kind"],
                            song_dmca=song.get("dmca", False),
                        )

            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_KH1,
                game_music_path=extracted_data_path / "kh1" / "remastered" / "amusic",
                game_music_list=music.kh1_music_list,
            )
            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_RECOM,
                game_music_path=extracted_data_path / "recom" / "STREAM" / "0001",
                game_music_list=music.recom_music_list,
            )
            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_BBS,
                game_music_path=extracted_data_path / "bbs" / "sound" / "win" / "bgm",
                game_music_list=music.bbs_music_list,
            )
            add_other_game_music(
                enabled_key=settingkey.MUSIC_RANDO_PC_INCLUDE_DDD,
                game_music_path=extracted_data_path
                / "kh3d"
                / "sound"
                / "jp"
                / "output"
                / "BGM",
                game_music_list=music.ddd_music_list,
            )

        return result

    @staticmethod
    def _get_music_assets(settings: SeedSettings) -> tuple[list[Asset], dict[str, str]]:
        music_rando_enabled = settings.get(settingkey.MUSIC_RANDO_ENABLED_PC)
        if not music_rando_enabled:
            return [], {}

        allow_duplicates = settings.get(settingkey.MUSIC_RANDO_PC_ALLOW_DUPLICATES)

        # Primary copy of the music. Songs are removed as they are chosen.
        music_files_by_categories: dict[str, list[Path]] = {}
        # Secondary copy of the music. Used to refill the lists once all are gone.
        backup_files_by_categories: dict[str, list[Path]] = {}

        music_files = CosmeticsMod._collect_music_files(settings)
        for category, song_list in music_files.items():
            main_list = song_list.copy()
            random.shuffle(main_list)
            music_files_by_categories[category] = main_list

            backup_files_by_categories[category] = song_list

        assets: list[Asset] = []
        replacements: dict[str, str] = {}

        music_list_file_path = CosmeticsMod.bootstrap_music_list_file()
        with open(music_list_file_path, encoding="utf-8") as music_list_file:
            music_metadata = json.load(music_list_file)
        random.shuffle(music_metadata)
        for info in music_metadata:
            filename = info["filename"]
            title = info["title"]
            types = [
                song_type.lower()
                for song_type in info["type"]
                if song_type.lower() in music_files_by_categories
            ]
            if len(types) > 0:
                random.shuffle(types)

                for chosen_type in types:
                    songs_for_chosen_type = music_files_by_categories[chosen_type]

                    # If there aren't any songs left for the type, attempt to refill the list if duplicates are allowed.
                    # If duplicates are not allowed and we run out of replacements, more will end up un-randomized.
                    if len(songs_for_chosen_type) == 0 and allow_duplicates:
                        refill_list = backup_files_by_categories[chosen_type].copy()
                        random.shuffle(refill_list)
                        music_files_by_categories[chosen_type] = refill_list
                        songs_for_chosen_type = refill_list

                    if len(songs_for_chosen_type) > 0:
                        chosen_song = str(songs_for_chosen_type.pop())

                        asset = {
                            "name": filename,
                            "vanilla_song": title,
                            "song_type": chosen_type,
                            "method": "copy",
                            "source": [{"name": chosen_song}],
                        }
                        assets.append(asset)

                        replacements[title] = chosen_song

                        break

        return assets, replacements

    @staticmethod
    def bootstrap_custom_visuals_folder(custom_visuals_path: Path):
        """Creates folders for each of the default custom visuals categories."""
        for folder in ["room-transition-images"]:
            (custom_visuals_path / folder).mkdir(exist_ok=True)

        end_pictures_path = custom_visuals_path / "ending-pictures"
        end_pictures_path.mkdir(exist_ok=True)

        item_pictures_path = custom_visuals_path / "item-pictures"
        item_pictures_path.mkdir(exist_ok=True)

        itempic_categories: list[str] = []
        for itempic in ItempicRandomizer.replaceable_itempics():
            itempic_categories.extend(itempic.types)
        for category in set(itempic_categories):
            (item_pictures_path / category.lower()).mkdir(exist_ok=True)

    @staticmethod
    def randomize_keyblades(
        seed_settings: SeedSettings,
    ) -> tuple[list[Asset], dict[str, str]]:
        """
        Randomizes keyblades, returning a list of assets to be added to the seed mod and a dictionary of which keyblade
        was replaced by which replacement.
        """
        return KeybladeRandomizer.randomize_keyblades(
            setting=seed_settings.get(settingkey.KEYBLADE_RANDO),
            include_effects=seed_settings.get(
                settingkey.KEYBLADE_RANDO_INCLUDE_EFFECTS
            ),
            allow_duplicate_replacement=seed_settings.get(
                settingkey.KEYBLADE_RANDO_ALLOW_DUPLICATES
            ),
        )

    @staticmethod
    def randomize_field2d(seed_settings: SeedSettings) -> list[Asset]:
        """Randomizes various field2d entries, returning a list of assets to be added to a mod."""
        assets: list[Asset] = []

        command_menu_choice = seed_settings.get(settingkey.COMMAND_MENU)
        assets.extend(
            CommandMenuRandomizer(command_menu_choice).randomize_command_menus()
        )

        transition_choice = seed_settings.get(settingkey.ROOM_TRANSITION_IMAGES)
        assets.extend(
            RoomTransitionImageRandomizer(
                transition_choice
            ).randomize_room_transitions()
        )

        return assets

    @staticmethod
    def randomize_itempics(seed_settings: SeedSettings) -> list[Asset]:
        """Randomizes various itempic entries, returning a list of assets to be added to a mod."""
        setting = seed_settings.get(settingkey.ITEMPIC_RANDO)
        return ItempicRandomizer.randomize_itempics(setting)

    @staticmethod
    def randomize_end_screen(seed_settings: SeedSettings) -> list[Asset]:
        """Randomizes the ending screen, returning a list of assets to be added to a mod."""
        setting = seed_settings.get(settingkey.ENDPIC_RANDO)
        return EndingPictureRandomizer.randomize_end_screen(setting)



from enum import Enum

from ...Class.newLocationClass import KH2Location
from ..configDict import locationType
from ..inventory import proof
from .graph import chest, LocationGraphBuilder, START_NODE


class CheckLocation(str, Enum):
    GardenOfAssemblageMap = "Garden of Assemblage Map"
    GoaLostIllusion = "GoA Lost Illusion"
    ProofOfNonexistence = "Proof of Nonexistence"


def make_graph(graph: LocationGraphBuilder):
    locations = [
        chest(i, f"Starting Item {i}", locationType.Critical) for i in range(1, 8)
    ]
    locations.append(chest(585, CheckLocation.GardenOfAssemblageMap, locationType.Free))
    locations.append(chest(586, CheckLocation.GoaLostIllusion, locationType.Free))
    locations.append(
        chest(
            590,
            CheckLocation.ProofOfNonexistence,
            locationType.Free,
            vanilla=proof.ProofOfNonexistence,
        )
    )

    graph.add_location(START_NODE, locations)


def donald_starting_items() -> list[KH2Location]:
    return [
        chest(i, f"Donald Starting Item {i}", locationType.Free) for i in range(1, 3)
    ]


def goofy_starting_items() -> list[KH2Location]:
    return [
        chest(i, f"Goofy Starting Item {i}", locationType.Free) for i in range(1, 3)
    ]



from dataclasses import dataclass, field
from typing import Optional

from ..configDict import itemType, locationType
from ..inventory.item import InventoryItem


@dataclass(frozen=True)
class StoryUnlock(InventoryItem):
    id: int
    name: str
    location: locationType
    visit_count: int
    type: itemType = field(init=False, default=itemType.STORYUNLOCK)


BattlefieldsOfWar = StoryUnlock(
    54, "Battlefields of War (Auron)", locationType.OC, visit_count=2
)
SwordOfTheAncestor = StoryUnlock(
    55, "Sword of the Ancestor (Mulan)", locationType.LoD, visit_count=2
)
BeastsClaw = StoryUnlock(59, "Beast's Claw (Beast)", locationType.BC, visit_count=2)
BoneFist = StoryUnlock(
    60, "Bone Fist (Jack Skellington)", locationType.HT, visit_count=2
)
ProudFang = StoryUnlock(61, "Proud Fang (Simba)", locationType.PL, visit_count=2)
SkillAndCrossbones = StoryUnlock(
    62, "Skill and Crossbones (Jack Sparrow)", locationType.PR, visit_count=2
)
Scimitar = StoryUnlock(72, "Scimitar (Aladdin)", locationType.Agrabah, visit_count=2)
WayToTheDawn = StoryUnlock(
    73, "Way to the Dawn (Riku)", locationType.TWTNW, visit_count=2
)
IdentityDisk = StoryUnlock(74, "Identity Disk (Tron)", locationType.SP, visit_count=2)
# TournamentPoster = StoryUnlock(365, "Tournament Poster")
MembershipCard = StoryUnlock(369, "Membership Card", locationType.HB, visit_count=2)
IceCream = StoryUnlock(375, "Ice Cream", locationType.TT, visit_count=3)
# TODO(zaktherobot): leaving out picture for now, may revisit if there's no confusion that picture doesn't unlock TT2 anymore
# Picture = StoryUnlock(376, "Picture")
NaminesSketches = StoryUnlock(368, "Namine's Sketches", locationType.STT, visit_count=1)
RoyalSummons = StoryUnlock(
    460, "Royal Summons", locationType.DC, visit_count=2
)  # Dummy 13


def all_story_unlocks() -> list[StoryUnlock]:
    return [
        BattlefieldsOfWar,
        SwordOfTheAncestor,
        BeastsClaw,
        BoneFist,
        ProudFang,
        SkillAndCrossbones,
        Scimitar,
        WayToTheDawn,
        IdentityDisk,
        MembershipCard,
        IceCream,
        NaminesSketches,
        RoyalSummons,
    ]


def all_individual_story_unlocks() -> list[StoryUnlock]:
    result: list[StoryUnlock] = []
    for unlock in all_story_unlocks():
        result.extend([unlock] * unlock.visit_count)
    return result


def story_unlock_for_location(location: locationType) -> Optional[StoryUnlock]:
    return next(
        (unlock for unlock in all_story_unlocks() if unlock.location == location), None
    )



from math import ceil, floor
from typing import Union

from ..Class.newLocationClass import KH2Location
from ..List.LvupStats import DreamWeaponOffsets
from ..List.NewLocationList import Locations
from ..List.configDict import (
    itemType,
    locationCategory,
    itemRarity,
    locationDepth,
    locationType,
    itemBias,
    itemDifficulty,
)
from ..List.location.graph import START_NODE
from .RandomizerSettings import RandomizerSettings


class SimplifiedWeightDistributions:
    def __init__(self, max_depth: int):
        self.weighting_function: dict[itemBias, list[int]] = {}

        weight_list_maker = lambda function_index: [
            max(
                1,
                floor(
                    10.0
                    / pow(
                        (max_depth + 2 - x) / (max_depth / 3.0),
                        function_index * 0.5 + 1,
                    )
                ),
            )
            for x in range(1, max_depth + 2)
        ]

        self.weighting_function[itemBias.VERY_EARLY] = weight_list_maker(1)[::-1]
        self.weighting_function[itemBias.EARLY] = weight_list_maker(0)[::-1]
        self.weighting_function[itemBias.SLIGHTLY_EARLY] = [2] * floor(
            (max_depth + 1) / 2
        ) + [1] * ceil((max_depth + 1) / 2)
        self.weighting_function[itemBias.NOBIAS] = [1] * (max_depth + 1)
        self.weighting_function[itemBias.SLIGHTLY_LATE] = [1] * floor(
            (max_depth + 1) / 2
        ) + [2] * ceil((max_depth + 1) / 2)
        self.weighting_function[itemBias.LATE] = weight_list_maker(0)
        self.weighting_function[itemBias.VERY_LATE] = weight_list_maker(1)
        self.weighting_function[itemBias.SUPER_LATE] = weight_list_maker(2)
        self.weighting_function[itemBias.NIGHTMARE] = weight_list_maker(3)

    def get_rarity_weightings(self):
        return self.weighting_function


class WeightDistributions:

    def __init__(self, max_depth: int):
        self.weighting_function: dict[itemDifficulty, dict[itemRarity, list[int]]] = {
            itemDifficulty.NORMAL: {}
        }
        for rarity in itemRarity:
            self.weighting_function[itemDifficulty.NORMAL][itemRarity(rarity)] = [1] * (
                max_depth + 1
            )

        self.weighting_function[itemDifficulty.SLIGHTLY_HARD] = {
            itemRarity.COMMON: [1] * (max_depth + 1),
            itemRarity.UNCOMMON: [1] * (max_depth + 1),
            itemRarity.RARE: [1] * floor((max_depth + 1) / 2)
            + [2] * ceil((max_depth + 1) / 2),
            itemRarity.MYTHIC: [1] * floor((max_depth + 1) / 2)
            + [2] * ceil((max_depth + 1) / 2),
        }
        self.weighting_function[itemDifficulty.SLIGHTLY_EASY] = {
            itemRarity.COMMON: [1] * (max_depth + 1),
            itemRarity.UNCOMMON: [1] * (max_depth + 1),
            itemRarity.RARE: [2] * floor((max_depth + 1) / 2)
            + [1] * ceil((max_depth + 1) / 2),
            itemRarity.MYTHIC: [2] * floor((max_depth + 1) / 2)
            + [1] * ceil((max_depth + 1) / 2),
        }

        easier_difficulties = [itemDifficulty.SUPEREASY, itemDifficulty.EASY]
        for diff_index, difficulty in enumerate(easier_difficulties):
            diff_index_modded = (1 - diff_index) - 0.25
            self.weighting_function[difficulty] = {
                itemRarity.COMMON: [
                    max(
                        1,
                        floor(
                            10.0
                            / pow(x / (max_depth / 3.0), diff_index_modded * 0.5 + 1)
                        ),
                    )
                    for x in range(max_depth + 1, 0, -1)
                ],
                itemRarity.UNCOMMON: [
                    max(
                        1,
                        floor(
                            3.0
                            / pow(x / (max_depth / 3.0), diff_index_modded * 0.5 + 1)
                        ),
                    )
                    for x in range(max_depth + 1, 0, -1)
                ],
                itemRarity.RARE: [
                    max(
                        1,
                        floor(
                            3.0
                            / pow(
                                (max_depth + 2 - x) / (max_depth / 3.0),
                                diff_index_modded * 0.5 + 1,
                            )
                        ),
                    )
                    for x in range(max_depth + 1, 0, -1)
                ],
                itemRarity.MYTHIC: [
                    max(
                        1,
                        floor(
                            10.0
                            / pow(
                                (max_depth + 2 - x) / (max_depth / 3.0),
                                diff_index_modded * 0.5 + 1,
                            )
                        ),
                    )
                    for x in range(max_depth + 1, 0, -1)
                ],
            }

        harder_difficulties = [
            itemDifficulty.HARD,
            itemDifficulty.VERYHARD,
            itemDifficulty.INSANE,
            itemDifficulty.NIGHTMARE,
        ]
        for diff_index, difficulty in enumerate(harder_difficulties):
            diff_index_modded = diff_index - 0.25
            self.weighting_function[difficulty] = {
                itemRarity.COMMON: [
                    max(
                        1,
                        floor(
                            10.0
                            / pow(x / (max_depth / 3.0), diff_index_modded * 0.5 + 1)
                        ),
                    )
                    for x in range(1, max_depth + 2)
                ],
                itemRarity.UNCOMMON: [
                    max(
                        1,
                        floor(
                            3.0
                            / pow(x / (max_depth / 3.0), diff_index_modded * 0.5 + 1)
                        ),
                    )
                    for x in range(1, max_depth + 2)
                ],
                itemRarity.RARE: [
                    max(
                        1,
                        floor(
                            3.0
                            / pow(
                                (max_depth + 2 - x) / (max_depth / 3.0),
                                diff_index_modded * 0.5 + 1,
                            )
                        ),
                    )
                    for x in range(1, max_depth + 2)
                ],
                itemRarity.MYTHIC: [
                    max(
                        1,
                        floor(
                            10.0
                            / pow(
                                (max_depth + 2 - x) / (max_depth / 3.0),
                                diff_index_modded * 0.5 + 1,
                            )
                        ),
                    )
                    for x in range(1, max_depth + 2)
                ],
            }

        # for diff_index,difficulty in enumerate(["Super Easy","Easy","Hard","Very Hard","Insane","Nightmare"]):
        #     print(self.weighting_function[difficulty][itemRarity.COMMON])
        #     print(self.weighting_function[difficulty][itemRarity.UNCOMMON])
        #     print(self.weighting_function[difficulty][itemRarity.RARE])
        #     print(self.weighting_function[difficulty][itemRarity.MYTHIC])
        #     print("--------------------")

    def get_rarity_weighting(self, difficulty: itemDifficulty):
        if difficulty not in self.weighting_function:
            raise ValueError(f"Unknown item placement difficulty {difficulty}")
        return self.weighting_function[difficulty]


class LocationWeights:

    def __init__(
        self,
        settings: RandomizerSettings,
        locations: Locations,
        reverse_locations: Locations,
    ):
        self.regular_rando = settings.regular_rando
        self.reverse_rando = settings.reverse_rando
        self.split_levels = settings.split_levels
        self.level_offsets = DreamWeaponOffsets()
        self.max_level = settings.max_level_checks
        self.level_depths: dict[int, int] = {}

        # Regular
        # -------------------------------
        self.location_depths: dict[KH2Location, int] = {}
        hops = locations.location_graph.get_hops(START_NODE)
        location_type_maxes: dict[locationType, int] = {}
        max_hops = (
            17  # hard coding max hops to unify depths between reverse and regular rando
        )
        # for hop in hops:
        #     max_hops = max(max_hops,hop[1])
        for node, distance in hops:
            for loc in locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if primary_type not in location_type_maxes:
                    location_type_maxes[primary_type] = distance
                else:
                    location_type_maxes[primary_type] = max(
                        distance, location_type_maxes[primary_type]
                    )

        for node, distance in hops:
            for loc in locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if location_type_maxes[primary_type] != 0:
                    scaled_depth = floor(
                        (distance * 1.0 / location_type_maxes[primary_type]) * max_hops
                    )
                else:
                    scaled_depth = distance
                self.location_depths[loc] = scaled_depth
                if loc.LocationCategory is locationCategory.LEVEL:
                    self.level_depths[loc.LocationId] = scaled_depth

        # Reverse
        # -------------------------------
        self.reverse_location_depths: dict[KH2Location, int] = {}
        hops = reverse_locations.location_graph.get_hops(START_NODE)
        reverse_location_type_maxes: dict[locationType, int] = {}
        # max_hops = 0
        # for hop in hops:
        #     max_hops = max(max_hops,hop[1])
        for node, distance in hops:
            for loc in reverse_locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if primary_type not in reverse_location_type_maxes:
                    reverse_location_type_maxes[primary_type] = distance
                else:
                    reverse_location_type_maxes[primary_type] = max(
                        distance, reverse_location_type_maxes[primary_type]
                    )

        for node, distance in hops:
            for loc in reverse_locations.locations_for_node(node):
                primary_type = loc.LocationTypes[0]
                if reverse_location_type_maxes[primary_type] != 0:
                    scaled_depth = floor(
                        (distance * 1.0 / reverse_location_type_maxes[primary_type])
                        * max_hops
                    )
                else:
                    scaled_depth = distance
                self.reverse_location_depths[loc] = scaled_depth

        self.base_weights = SimplifiedWeightDistributions(
            max_hops
        ).get_rarity_weightings()
        self.weights = {}
        # default to normal weighting
        for item_type in itemType:
            self.weights[item_type] = self.base_weights[itemBias.NOBIAS]

        self.weights[itemType.FORM] = self.base_weights[settings.form_weights]
        self.weights[itemType.STORYUNLOCK] = self.base_weights[settings.unlock_weights]

        self.weights[itemType.FIRE] = self.base_weights[settings.magic_weights]
        self.weights[itemType.BLIZZARD] = self.base_weights[settings.magic_weights]
        self.weights[itemType.THUNDER] = self.base_weights[settings.magic_weights]
        self.weights[itemType.CURE] = self.base_weights[settings.magic_weights]
        self.weights[itemType.MAGNET] = self.base_weights[settings.magic_weights]
        self.weights[itemType.REFLECT] = self.base_weights[settings.magic_weights]

        self.weights[itemType.TORN_PAGE] = self.base_weights[settings.page_weights]

        self.weights[itemType.SUMMON] = self.base_weights[settings.summon_weights]

        self.weights[itemType.REPORT] = self.base_weights[settings.report_weights]

        self.weights[itemType.PROOF_OF_PEACE] = self.base_weights[
            settings.proof_weights
        ]
        self.weights[itemType.PROOF_OF_CONNECTION] = self.base_weights[
            settings.proof_weights
        ]
        self.weights[itemType.PROOF_OF_NONEXISTENCE] = self.base_weights[
            settings.proof_weights
        ]

        self.weights[itemType.PROMISE_CHARM] = self.base_weights[
            settings.promise_charm_weights
        ]

    def get_depth(self, location: KH2Location) -> Union[int, tuple[int, int]]:
        """
        Returns either a single depth (for levels or regular or reverse rando) or a tuple of
        (regular depth, reverse depth) if both regular and reverse rando.
        """
        if location.LocationCategory is locationCategory.LEVEL:
            loc_id = location.LocationId
            if self.split_levels:
                sword_depth = self.level_depths[loc_id]
                shield_depth = self.level_depths[
                    self.level_offsets.get_shield_level(self.max_level, loc_id)
                ]
                staff_depth = self.level_depths[
                    self.level_offsets.get_staff_level(self.max_level, loc_id)
                ]
                return (sword_depth + shield_depth + staff_depth) // 3
            else:
                return self.level_depths[loc_id]
        else:
            if self.regular_rando and self.reverse_rando:
                return (
                    self.location_depths[location],
                    self.reverse_location_depths[location],
                )
            elif self.regular_rando:
                return self.location_depths[location]
            else:
                return self.reverse_location_depths[location]

    def get_weight(self, rarity: itemRarity, location: KH2Location) -> int:
        raise AssertionError("Deprecated weighting function")
        """ Returns the weight that should be used for an item of the given rarity at the given location. """
        depth_or_depths = self.get_depth(location)
        rarity_weights = self.weights[rarity]
        if isinstance(depth_or_depths, int):
            return rarity_weights[depth_or_depths]
        elif isinstance(depth_or_depths, tuple):
            regular_weight = rarity_weights[depth_or_depths[0]]
            reverse_weight = rarity_weights[depth_or_depths[1]]
            return (regular_weight + reverse_weight) // 2

    def get_weight(self, item_type: itemType, location: KH2Location) -> int:
        """Returns the weight that should be used for an item of the given type at the given location."""
        depth_or_depths = self.get_depth(location)
        rarity_weights = self.weights[item_type]
        if isinstance(depth_or_depths, int):
            return rarity_weights[depth_or_depths]
        elif isinstance(depth_or_depths, tuple):
            regular_weight = rarity_weights[depth_or_depths[0]]
            reverse_weight = rarity_weights[depth_or_depths[1]]
            return (regular_weight + reverse_weight) // 2



import argparse

from .seed_gen import generator

# CLI arguments & docs
parser = argparse.ArgumentParser(
    description="A command line utility for generating KH2 Randomizer seed strings from a given preset"
)
parser.add_argument(
    "preset",
    help='preset from which to generate seed string (default: "League Summer 2025")',
    nargs="?",
    default="League Summer 2025",
)
args = parser.parse_args()
requested_preset = args.preset


def app():
    seed_info = generator.make_random_seed_from_preset_name(requested_preset)
    print(seed_info.generator_string)
    print(seed_info.hash_icons)


if __name__ == "__main__":
    app()



from enum import Enum

from ..configDict import locationType, itemType
from ..inventory import magic, keyblade, ability, misc
from .graph import (
    DefaultLogicGraph,
    RequirementEdge,
    chest,
    popup,
    hybrid_bonus,
    stat_bonus,
    LocationGraphBuilder,
    START_NODE,
)
from ...Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    Graveyard = "Graveyard"
    GraveyardChests = "Graveyard Chests"
    FinklesteinsLab = "Finklestein's Lab"
    FinklesteinsLabChests = "Finklestein's Lab Chests"
    TownSquare = "Town Square"
    TownSquareChests = "Town Square Chests"
    Hinterlands = "Hinterlands"
    HinterlandsChests = "Hinterlands Chests"
    CandyCaneLane = "Candy Cane Lane"
    CandyCaneLaneChests = "Candy Cane Lane Chests"
    SantasHouse = "Santa's House"
    SantasHouseChests = "Santa's House Chests"
    PrisonKeeper = "Prison Keeper"
    OogieBoogie = "Oogie Boogie"
    LockShockBarrel = "Lock, Shock, and Barrel"
    Presents = "Presents"
    DecoyPresentMinigame = "Decoy Present Minigame"
    Experiment = "Experiment"
    Vexen = "AS Vexen"
    DataVexen = "Data Vexen"


class CheckLocation(str, Enum):
    GraveyardMythrilShard = "Graveyard Mythril Shard"
    GraveyardSerenityGem = "Graveyard Serenity Gem"
    FinklesteinsLabHalloweenTownMap = "Finklestein's Lab Halloween Town Map"
    TownSquareMythrilStone = "Town Square Mythril Stone"
    TownSquareEnergyShard = "Town Square Energy Shard"
    HinterlandsLightningShard = "Hinterlands Lightning Shard"
    HinterlandsMythrilStone = "Hinterlands Mythril Stone"
    HinterlandsApBoost = "Hinterlands AP Boost"
    CandyCaneLaneMegaPotion = "Candy Cane Lane Mega-Potion"
    CandyCaneLaneMythrilGem = "Candy Cane Lane Mythril Gem"
    CandyCaneLaneLightningStone = "Candy Cane Lane Lightning Stone"
    CandyCaneLaneMythrilStone = "Candy Cane Lane Mythril Stone"
    SantasHouseChristmasTownMap = "Santa's House Christmas Town Map"
    SantasHouseApBoost = "Santa's House AP Boost"
    PrisonKeeper = "Prison Keeper"
    OogieBoogie = "Oogie Boogie"
    OogieMagnetElement = "Oogie Boogie Magnet Element"
    LockShockBarrel = "Lock, Shock, and Barrel"
    Present = "Present"
    DecoyPresents = "Decoy Presents"
    Experiment = "Experiment"
    DecisivePumpkin = "Decisive Pumpkin"
    VexenBonus = "Vexen Bonus"
    VexenRoadToDiscovery = "Vexen (AS) Road to Discovery"
    DataVexen = "Vexen (Data) Lost Illusion"


class HTLogicGraph(DefaultLogicGraph):
    def __init__(self, reverse_rando, keyblade_unlocks):
        DefaultLogicGraph.__init__(self, NodeId)
        keyblade_lambda = (
            lambda inv: not keyblade_unlocks
            or ItemPlacementHelpers.need_ht_keyblade(inv)
        )
        self.logic[NodeId.Graveyard][NodeId.GraveyardChests] = keyblade_lambda
        self.logic[NodeId.FinklesteinsLab][
            NodeId.FinklesteinsLabChests
        ] = keyblade_lambda
        self.logic[NodeId.TownSquare][NodeId.TownSquareChests] = keyblade_lambda
        self.logic[NodeId.Hinterlands][NodeId.HinterlandsChests] = keyblade_lambda
        self.logic[NodeId.CandyCaneLane][NodeId.CandyCaneLaneChests] = keyblade_lambda
        self.logic[NodeId.SantasHouse][NodeId.SantasHouseChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.Graveyard] = ItemPlacementHelpers.ht1_check
            self.logic[NodeId.OogieBoogie][
                NodeId.LockShockBarrel
            ] = ItemPlacementHelpers.ht2_check
        else:
            self.logic[START_NODE][NodeId.SantasHouse] = ItemPlacementHelpers.ht1_check
            self.logic[NodeId.Vexen][
                NodeId.FinklesteinsLab
            ] = ItemPlacementHelpers.ht2_check


def make_graph(graph: LocationGraphBuilder):
    ht = locationType.HT
    ht_logic = HTLogicGraph(graph.reverse_rando, graph.keyblades_unlock_chests)
    graph.add_logic(ht_logic)

    finklesteins_lab_chests = graph.add_location(
        NodeId.FinklesteinsLabChests,
        [
            chest(211, CheckLocation.FinklesteinsLabHalloweenTownMap, ht),
        ],
    )
    finklesteins_lab = graph.add_location(NodeId.FinklesteinsLab, [])
    graveyard_chests = graph.add_location(
        NodeId.GraveyardChests,
        [
            chest(53, CheckLocation.GraveyardMythrilShard, ht),
            chest(212, CheckLocation.GraveyardSerenityGem, ht),
        ],
    )
    graveyard = graph.add_location(NodeId.Graveyard, [])
    town_square_chests = graph.add_location(
        NodeId.TownSquareChests,
        [
            chest(209, CheckLocation.TownSquareMythrilStone, ht),
            chest(210, CheckLocation.TownSquareEnergyShard, ht),
        ],
    )
    town_square = graph.add_location(NodeId.TownSquare, [])
    hinterlands_chests = graph.add_location(
        NodeId.HinterlandsChests,
        [
            chest(54, CheckLocation.HinterlandsLightningShard, ht),
            chest(213, CheckLocation.HinterlandsMythrilStone, ht),
            chest(214, CheckLocation.HinterlandsApBoost, ht),
        ],
    )
    hinterlands = graph.add_location(NodeId.Hinterlands, [])
    candy_cane_lane_chests = graph.add_location(
        NodeId.CandyCaneLaneChests,
        [
            chest(55, CheckLocation.CandyCaneLaneMegaPotion, ht),
            chest(56, CheckLocation.CandyCaneLaneMythrilGem, ht),
            chest(216, CheckLocation.CandyCaneLaneLightningStone, ht),
            chest(217, CheckLocation.CandyCaneLaneMythrilStone, ht),
        ],
    )
    candy_cane_lane = graph.add_location(NodeId.CandyCaneLane, [])
    santas_house_chests = graph.add_location(
        NodeId.SantasHouseChests,
        [
            chest(
                57,
                CheckLocation.SantasHouseChristmasTownMap,
                ht,
                invalid_checks=[itemType.GAUGE],
            ),
            chest(
                58,
                CheckLocation.SantasHouseApBoost,
                ht,
                invalid_checks=[itemType.GAUGE],
            ),
        ],
    )
    santas_house = graph.add_location(NodeId.SantasHouse, [])
    prison_keeper = graph.add_location(
        NodeId.PrisonKeeper,
        [
            hybrid_bonus(18, CheckLocation.PrisonKeeper, ht, vanilla=ability.FlashStep),
        ],
    )
    oogie_boogie = graph.add_location(
        NodeId.OogieBoogie,
        [
            stat_bonus(19, CheckLocation.OogieBoogie, ht),
            popup(301, CheckLocation.OogieMagnetElement, ht, vanilla=magic.Magnet),
        ],
    )
    lock_shock_barrel = graph.add_location(
        NodeId.LockShockBarrel,
        [
            stat_bonus(40, CheckLocation.LockShockBarrel, ht),
        ],
    )
    presents = graph.add_location(
        NodeId.Presents,
        [
            popup(297, CheckLocation.Present, ht, vanilla=misc.Present),
        ],
    )
    decoy_present_minigame = graph.add_location(
        NodeId.DecoyPresentMinigame,
        [
            popup(298, CheckLocation.DecoyPresents, ht, vanilla=misc.DecoyPresents),
        ],
    )
    experiment = graph.add_location(
        NodeId.Experiment,
        [
            stat_bonus(20, CheckLocation.Experiment, ht),
            popup(
                275, CheckLocation.DecisivePumpkin, ht, vanilla=keyblade.DecisivePumpkin
            ),
        ],
    )
    vexen = graph.add_location(
        NodeId.Vexen,
        [
            stat_bonus(64, CheckLocation.VexenBonus, [ht, locationType.AS]),
            popup(544, CheckLocation.VexenRoadToDiscovery, [ht, locationType.AS]),
        ],
    )
    data_vexen = graph.add_location(
        NodeId.DataVexen,
        [
            popup(549, CheckLocation.DataVexen, [ht, locationType.DataOrg]),
        ],
    )

    graph.register_superboss(data_vexen)

    graph.add_edge(finklesteins_lab, finklesteins_lab_chests)
    graph.add_edge(graveyard, graveyard_chests)
    graph.add_edge(town_square, town_square_chests)
    graph.add_edge(hinterlands, hinterlands_chests)
    graph.add_edge(candy_cane_lane, candy_cane_lane_chests)
    graph.add_edge(santas_house, santas_house_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, graveyard)
        graph.add_edge(graveyard, finklesteins_lab)
        graph.add_edge(finklesteins_lab, town_square)
        graph.add_edge(town_square, hinterlands)
        graph.add_edge(hinterlands, candy_cane_lane, RequirementEdge(battle=True))
        graph.add_edge(candy_cane_lane, santas_house)
        graph.add_edge(santas_house, prison_keeper, RequirementEdge(battle=True))
        graph.add_edge(prison_keeper, oogie_boogie, RequirementEdge(battle=True))
        graph.add_edge(oogie_boogie, lock_shock_barrel, RequirementEdge(battle=True))
        graph.add_edge(lock_shock_barrel, presents, RequirementEdge(battle=True))
        graph.add_edge(presents, decoy_present_minigame)
        graph.add_edge(decoy_present_minigame, experiment, RequirementEdge(battle=True))
        graph.add_edge(experiment, vexen, RequirementEdge(battle=True))
        graph.add_edge(vexen, data_vexen)
        graph.register_first_boss(oogie_boogie)
        graph.register_last_story_boss(experiment)
        graph.register_superboss(vexen)
    else:
        graph.add_edge(START_NODE, santas_house)
        graph.add_edge(santas_house, candy_cane_lane)
        graph.add_edge(candy_cane_lane, hinterlands)
        graph.add_edge(hinterlands, graveyard)
        graph.add_edge(graveyard, town_square)
        graph.add_edge(santas_house, lock_shock_barrel, RequirementEdge(battle=True))
        graph.add_edge(lock_shock_barrel, presents, RequirementEdge(battle=True))
        graph.add_edge(presents, decoy_present_minigame)
        graph.add_edge(decoy_present_minigame, experiment, RequirementEdge(battle=True))
        graph.add_edge(experiment, vexen, RequirementEdge(battle=True))
        graph.add_edge(vexen, finklesteins_lab)
        graph.add_edge(finklesteins_lab, prison_keeper, RequirementEdge(battle=True))
        graph.add_edge(prison_keeper, oogie_boogie, RequirementEdge(battle=True))
        graph.add_edge(oogie_boogie, data_vexen, RequirementEdge(battle=True))
        graph.register_first_boss(experiment)
        graph.register_last_story_boss(oogie_boogie)



import random
from typing import Callable

from ..Class.exceptions import GeneratorException, SettingsException
from ..Class.itemClass import KH2Item
from ..List.configDict import (
    LevelUpStatBonus,
    StartingMovementOption,
    AbilityPoolOption,
    StartingVisitMode,
    locationType,
)
from ..List.inventory import growth, ability, storyunlock
from ..List.inventory.growth import GrowthAbility, GrowthType
from ..List.inventory.storyunlock import StoryUnlock


class SeedModifier:
    @staticmethod
    def ability_list_modifier(
        option: AbilityPoolOption,
    ) -> Callable[[list[KH2Item], list[KH2Item]], list[KH2Item]]:
        if option == AbilityPoolOption.DEFAULT:
            return SeedModifier.default_ability_pool
        elif option == AbilityPoolOption.RANDOMIZE:
            return SeedModifier.random_ability_pool
        elif option == AbilityPoolOption.RANDOMIZE_SUPPORT:
            return SeedModifier.random_support_ability_pool
        elif option == AbilityPoolOption.RANDOMIZE_STACKABLE:
            return SeedModifier.random_stackable_ability_pool
        else:
            raise SettingsException("Invalid ability pool option")

    @staticmethod
    def default_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        return action + support

    @staticmethod
    def random_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        randomizable_list = action + support
        randomizable_dict: dict[str, KH2Item] = {i.Name: i for i in randomizable_list}
        possible_abilities = list(
            set(
                i.Name
                for i in randomizable_list
                if i.Name not in [ability.SecondChance.name, ability.OnceMore.name]
            )
        )
        possible_abilities.sort()
        random_ability_pool = []
        for _ in range(len(randomizable_list) - 2):
            choice: str = random.choice(possible_abilities)
            random_ability_pool.append(randomizable_dict[choice])
            # Limit only 1 of each action ability in the pool, to make it more interesting
            if choice in [i.Name for i in action]:
                possible_abilities.remove(choice)

        # Make sure there is one OM and one SC so the tracker behaves
        random_ability_pool.append(randomizable_dict[ability.SecondChance.name])
        random_ability_pool.append(randomizable_dict[ability.OnceMore.name])
        return random_ability_pool

    @staticmethod
    def random_support_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        randomizable_list = support
        randomizable_dict: dict[str, KH2Item] = {i.Name: i for i in randomizable_list}
        possible_abilities = list(
            set(
                i.Name
                for i in randomizable_list
                if i.Name not in [ability.SecondChance.name, ability.OnceMore.name]
            )
        )
        possible_abilities.sort()
        random_ability_pool = []
        for _ in range(len(randomizable_list) - 2):
            choice: str = random.choice(possible_abilities)
            random_ability_pool.append(randomizable_dict[choice])

        # Make sure there is one OM and one SC so the tracker behaves
        random_ability_pool.append(randomizable_dict[ability.SecondChance.name])
        random_ability_pool.append(randomizable_dict[ability.OnceMore.name])
        return random_ability_pool + action

    @staticmethod
    def random_stackable_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        stackable_abilities = [
            ability.ComboPlus.name,
            ability.AirComboPlus.name,
            ability.ComboBoost.name,
            ability.AirComboBoost.name,
            ability.ReactionBoost.name,
            ability.FinishingPlus.name,
            ability.NegativeCombo.name,
            ability.BerserkCharge.name,
            ability.DamageDrive.name,
            ability.DriveBoost.name,
            ability.FormBoost.name,
            ability.SummonBoost.name,
            ability.ExperienceBoost.name,
            ability.Draw.name,
            ability.Jackpot.name,
            ability.LuckyLucky.name,
            ability.DriveConverter.name,
            ability.FireBoost.name,
            ability.BlizzardBoost.name,
            ability.ThunderBoost.name,
            ability.ItemBoost.name,
            ability.MpRage.name,
            ability.MpHaste.name,
            ability.MpHastera.name,
            ability.MpHastega.name,
            ability.Defender.name,
            ability.DamageControl.name,
            ability.CombinationBoost.name,
        ]
        ability_list = support + action
        ability_dict: dict[str, KH2Item] = {i.Name: i for i in ability_list}
        unique_abilities = list(set(i.Name for i in ability_list))
        unique_abilities.sort()
        random_ability_pool = []
        for unique_ability in unique_abilities:
            random_ability_pool.append(ability_dict[unique_ability])
        for _ in range(len(ability_list) - len(unique_abilities)):
            choice = random.choice(stackable_abilities)
            random_ability_pool.append(ability_dict[choice])

        return random_ability_pool

    @staticmethod
    def level_up_stat_pool(glass_cannon: bool) -> list[tuple[LevelUpStatBonus, int]]:
        stat_pool = SeedModifier.level_up_stat_pool_weighted(
            def_rate=0 if glass_cannon else 25
        )
        return [(s[0], s[1]) for s in stat_pool]

    @staticmethod
    def level_up_stat_pool_weighted(
        str_rate: int = 25,
        mag_rate: int = 25,
        def_rate: int = 25,
        ap_rate: int = 25,
    ) -> list[tuple[LevelUpStatBonus, int, int]]:
        valid_stat_list = [
            (LevelUpStatBonus.STRENGTH, 2, str_rate),
            (LevelUpStatBonus.MAGIC, 2, mag_rate),
            (LevelUpStatBonus.DEFENSE, 1, def_rate),
            (LevelUpStatBonus.AP, 2, ap_rate),
        ]
        # remove any stats with zero weight
        valid_stat_list = [
            stat_config for stat_config in valid_stat_list if stat_config[2] > 0
        ]
        if len(valid_stat_list) < 2:
            raise GeneratorException("Not enough valid stats for level ups")

        # rescale the weights to sum to 100
        total_sum = sum([stat_config[2] for stat_config in valid_stat_list])
        valid_stat_list = [
            (stat_config[0], stat_config[1], (100 * stat_config[2]) // total_sum)
            for stat_config in valid_stat_list
        ]
        new_total_sum = sum([stat_config[2] for stat_config in valid_stat_list])
        diff_adjustment = 0
        while new_total_sum + diff_adjustment < 100:
            old_tuple = valid_stat_list[diff_adjustment % len(valid_stat_list)]
            valid_stat_list[diff_adjustment % len(valid_stat_list)] = (
                old_tuple[0],
                old_tuple[1],
                old_tuple[2] + 1,
            )
            diff_adjustment += 1

        return valid_stat_list

    @staticmethod
    def starting_growth(option: StartingMovementOption) -> list[GrowthAbility]:
        if option == StartingMovementOption.DISABLED:
            return []
        elif option == StartingMovementOption.LEVEL_1:
            return growth.all_growth_to_level(1)
        elif option == StartingMovementOption.LEVEL_2:
            return growth.all_growth_to_level(2)
        elif option == StartingMovementOption.LEVEL_3:
            return growth.all_growth_to_level(3)
        elif option == StartingMovementOption.LEVEL_4:
            return growth.all_growth_to_level(4)
        elif option == StartingMovementOption.RANDOM_3:
            return SeedModifier._random_growth(3)
        elif option == StartingMovementOption.RANDOM_5:
            return SeedModifier._random_growth(5)
        elif option == StartingMovementOption.RANDOM_7:
            return SeedModifier._random_growth(7)
        elif option == StartingMovementOption.RANDOM_9:
            return SeedModifier._random_growth(9)
        else:
            raise GeneratorException(f"Unknown option {option}")

    @staticmethod
    def _random_growth(num_growth_abilities: int) -> list[GrowthAbility]:
        options = [growth_ability.growth_type for growth_ability in growth.all_growth()]
        levels_by_growth_type: dict[GrowthType, int] = {}
        for chosen_growth_type in random.sample(options, k=num_growth_abilities):
            if chosen_growth_type in levels_by_growth_type:
                levels_by_growth_type[chosen_growth_type] = (
                    levels_by_growth_type[chosen_growth_type] + 1
                )
            else:
                levels_by_growth_type[chosen_growth_type] = 1

        random_growth = []
        for growth_type, level in levels_by_growth_type.items():
            random_growth.extend(growth.growth_to_level(level, growth_type))
        return random_growth

    @staticmethod
    def starting_unlocks(
        mode: StartingVisitMode,
        random_range: tuple[int, int],
        specific_unlocks: dict[locationType, int],
    ) -> list[StoryUnlock]:
        if mode is StartingVisitMode.ALL:
            return storyunlock.all_individual_story_unlocks()
        elif mode is StartingVisitMode.FIRST:
            return storyunlock.all_story_unlocks()
        elif mode is StartingVisitMode.NONE:
            return []
        elif mode is StartingVisitMode.RANDOM:
            random_min, random_max = random_range
            random_count = random.randint(random_min, random_max)
            return random.sample(
                storyunlock.all_individual_story_unlocks(), k=random_count
            )
        elif mode is StartingVisitMode.SPECIFIC:
            result: list[StoryUnlock] = []
            for location, count in specific_unlocks.items():
                result.extend([storyunlock.story_unlock_for_location(location)] * count)
            return result
        elif mode is StartingVisitMode.CUSTOM:
            result: list[StoryUnlock] = []
            random_pool: list[StoryUnlock] = []
            for unlock in storyunlock.all_story_unlocks():
                specific_count = specific_unlocks[unlock.location]
                result.extend([unlock] * specific_count)
                random_pool.extend([unlock] * (unlock.visit_count - specific_count))
            random_min, random_max = random_range
            random_count = random.randint(random_min, random_max)
            result.extend(random.sample(random_pool, k=random_count))
            return result
        else:
            raise GeneratorException(f"Unknown mode {mode}")



import os
import sys

if not __package__:
    # Make CLI runnable from source tree with
    #    python src/package
    package_source_path = os.path.dirname(os.path.dirname(__file__))
    sys.path.insert(0, package_source_path)

if __name__ == "__main__":
    from kh2rando.seed_gen_cli import app

    app()
